#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 13 11:46:16 2017

@author: thiago
"""

class BaseDados:
    
    class _valor_chave:
        def __init__(self, chave, valor = None):
            self.chave = chave
            if valor:
                self.valor = valor
        def __lt__(self, outro):
            return self.chave < outro.chave
    
    def __init__(self):
        self.arvore = NoArvoreBinaria.ArvoreVazia
        
    def __getitem__(self, chave):
        return self.arvore.busca_no(self._valor_chave(chave)).val().valor
    
    def __setitem__(self, chave, valor):
        self.arvore = self.arvore.modifica(self._valor_chave(chave, valor))

class NoArvoreBinaria:
    # Esta classe funciona como "sentinela"
    class _NoNulo:
        _it_empty = iter([])
        def insere(self, x):
            return NoArvoreBinaria(x)
        def remove(self, x):
            raise Exception("Elemento nao existe")
        def __contains__(self, x):
            return False
        def __bool__(self):
            return False
        def __len__(self):
            return 0
        def __iter__(self):
            return self._it_empty
        def __getitem__(self, chave):
            raise KeyError
        def modifica(self, chave):
            return self.insere(chave)
        def busca_no(self, chave):
            raise KeyError
    
    ArvoreVazia = _NoNulo()
    
    def __init__(self, x, esquerda = ArvoreVazia, direita = ArvoreVazia):
        self._x = x
        self._e = esquerda
        self._d = direita
    
    def filhos(self):
        yield self._e
        yield self._d
    
    def e(self):
        return self._e
    
    def d(self):
        return self._d
        
    def val(self):
        return self._x
    
    def remove(self, x):
        if x > self._x:
            self._d = self._d.remove(x)
            return self
        elif x < self._x:
            self._e = self._e.remove(x)
            return self
        if not self._e:
            return self._d
        if not self._d:
            return self._e
        # Ambos os braços são não-nulos:
        #  Procura pelo menor elemento da arvore direita
        a = self._d
        while a._e: a = a._e
        # Troca-o com o elemento atual
        #  Note que aqui a estrutura está temporariamente
        #  quebrada
        a._x, self._x = self._x, a._x
        self._d = self._d.remove(x)
        return self
    
    def __iter__(self):
        yield self.val()
        for n in iter(self._e):
            yield n
        for n in iter(self._d):
            yield n
    
    def busca_no(self, chave):
        if chave < self._x:
            return self._d.busca_no(chave)
        elif self._x < chave:
            return self._e.busca_no(chave)
        return self
        
    def __getitem__(self, chave):
        if chave < self._x:
            return self._d[chave]
        elif self._x < chave:
            return self._e[chave]
        return self.val()
    
    def modifica(self, chave):
        if chave < self._x:
            self._d = self._d.modifica(chave)
        if self._x < chave:
            self._e = self._e.modifica(chave)
        self._x = chave
        return self
    
    def __contains__(self, x):
        if x > self._x:
            return x in self._d
        elif x < self._x:
            return x in self._e
        return True

    def __len__(self):
        """ Retorna o tamanho total da arvore"""
        return 1 + len(self._e) + len(self._d)

    def insere(self, x):
        if x>self._x:
            self._d = self._d.insere(x)
        elif x<self._x:
            self._e = self._e.insere(x)
        else:
            raise Exception("Elemento ja existe")
        return self
    
    def max_val(self):
        """ Retorna o maior valor da árvore"""
        while self._d:
            self = self._d
        return self.val()
    
    def min_val(self):
        """ Retorna o menor valor da árvore"""
        while self._e:
            self = self._e
        return self.val()    

def mostra_exp(arvore):
    s = ""
    if(arvore.e()):
        s += "("
        s += mostra_exp(arvore.e())
    s += str(arvore.val())
    if(arvore.d()):
        s += mostra_exp(arvore.d())
        s += ")"
    return s

def mostra_ant(arvore):
    s = ""
    s += str(arvore.val())
    if arvore.e():
        s += mostra_ant(arvore.e())
    if arvore.d():
        s += mostra_ant(arvore.d())
    return s

def mostra_int(arvore):
    s = ""
    if arvore.e():
        s += mostra_int(arvore.e())
    s += str(arvore.val())
    if arvore.d():
        s += mostra_int(arvore.d())
    return s

def mostra_pos(arvore):
    s = ""
    if arvore.e():
        s += mostra_pos(arvore.e())
    if arvore.d():
        s += mostra_pos(arvore.d())
    s += str(arvore.val())

    return s

class NoArvore:
    def __init__(self, x, filhos=[]):
        self._x = x
        self._filhos = filhos
    
    def filhos(self):
        return self._filhos
    
    def val(self):
        return self._x
     
